from pandas import json_normalize

from .helpers import generate_dualcase_string
from .discover import Discover
from .helpers import make_discover_request


class DatasetRepo(Discover):
    def __init__(self, endpoint):
        self.endpoint = endpoint

    def get_n_datasets_for_field(self,
                                 field,
                                 value,
                                 regexp,
                                 n=Discover.DEFAULT_PAGE_SIZE,
                                 cursor=None,
                                 sort_by=Discover.DEFAULT_SORT_ORDER):
        """Fetch dataset entries that match against the `value` for a `field`.
        Optionally a regular expression (`regexp`) can be provided instead of a
        `value`.

        Args:
            field (str):
                Name of the field whose values are to be matched against.
            value (str):
                Value to be matched.
            regexp (str):
                A regular expression in case more flexibility is desired rather
                than a plain `value`.
            n (int, optional):
                Number of matches to fetch. Defaults to
                `Discover.DEFAULT_PAGE_SIZE`.
            cursor (list, optional):
                A cursor from where to begin searching. Can be obtained from
                calls to this method. Defaults to None, in case search needs to
                start at the beginning.
            sort_by (list):
                A list of dictionaries which will be used to order responses
                for queries. Defaults to sorting based on combination of score
                (descending) and ID (ascending).

                Each dictionary is a key-value pair, where the keys are fields
                to be sorted on and the values may contain order, mode, etc. as
                documented here:
                https://www.elastic.co/guide/en/elasticsearch/reference/master/sort-search-results.html

        Returns:
            A list of hits and a cursor. The list contains the `n` hits as
            rows. The cursor can be used to fetch the next set of items from
            where this one ended.
        """
        payload = {}
        if (field == ""):
            payload = {
                "query": {
                    "multi_match": {
                        "query": value,
                        "fields": []
                    }
                },
            }
        else:
            payload = {
                "query": {"match": {field: value}},
            }

        if regexp:
            payload = {
                'query': {
                    'regexp': {
                        f'{field}.keyword': {
                            'value': f'.*{generate_dualcase_string(value)}.*',
                            'flags': 'ALL',
                            'max_determinized_states': 10000,
                            'rewrite': 'constant_score'
                        }
                    }
                },
            }

        if field is None or value is None:
            payload = None

        return self._execute_search_for_hits(payload, n, cursor, sort_by)

    def query_dataset_by_field(self,
                               field=None,
                               value=None,
                               regexp=False,
                               n=Discover.DEFAULT_PAGE_SIZE):
        print("Warning: this method has been deprecated. Please consider",
              "using 'get_n_datasets_for_field_as_df' instead.")

        hits, _ = self.get_n_datasets_for_field(field,
                                                value,
                                                regexp,
                                                n,
                                                None)
        search_df = json_normalize(hits)
        return search_df

    def get_n_datasets_for_field_as_df(self,
                                       field=None,
                                       value=None,
                                       regexp=False,
                                       n=Discover.DEFAULT_PAGE_SIZE,
                                       cursor=None,
                                       sort_by=Discover.DEFAULT_SORT_ORDER):
        """Obtain a dataframe of specified number of dataset matches against
        the `value` for a `field`. Optionally a regular expression (`regexp`)
        can be provided instead of a `value`.

        Args:
            field (str):
                Name of the field whose values are to be matched against.
            value (str):
                Value to be matched.
            regexp (str):
                A regular expression in case more flexibility is desired rather
                than a plain `value`.
            n (int, optional):
                Number of matches to fetch. Defaults to
                `Discover.DEFAULT_PAGE_SIZE`.
            cursor (list, optional):
                A cursor from where to begin searching. Can be obtained from
                calls to this method. Defaults to None, in case search needs to
                start at the beginning.
            sort_by (list):
                A list of dictionaries which will be used to order responses
                for queries. Defaults to sorting based on combination of score
                (descending) and ID (ascending).

                Each dictionary is a key-value pair, where the keys are fields
                to be sorted on and the values may contain order, mode, etc. as
                documented here:
                https://www.elastic.co/guide/en/elasticsearch/reference/master/sort-search-results.html

        Returns:
            A dataframe and a cursor. The dataframe contains the `n` hits as
            rows. The cursor can be used to fetch the next set of items from
            where this one ended.
        """
        hits, next_cursor = self.get_n_datasets_for_field(field,
                                                          value,
                                                          regexp,
                                                          n,
                                                          cursor,
                                                          sort_by)
        search_df = json_normalize(hits)
        return search_df, next_cursor

    def get_all_datasets_for_field_as_df(self,
                                         field=None,
                                         value=None,
                                         regexp=False,
                                         sort_by=Discover.DEFAULT_SORT_ORDER):
        """Obtain a dataframe of all dataset matches against the `value` for
        a `field`. Optionally a regular expression (`regexp`) can be provided
        instead of a `value`.

        Args:
            field (str):
                Name of the field whose values are to be matched against.
            value (str):
                Value to be matched.
            regexp (str):
                A regular expression in case more flexibility is desired rather
                than a plain `value`.
            sort_by (list):
                A list of dictionaries which will be used to order responses
                for queries. Defaults to sorting based on combination of score
                (descending) and ID (ascending).

                Each dictionary is a key-value pair, where the keys are fields
                to be sorted on and the values may contain order, mode, etc. as
                documented here:
                https://www.elastic.co/guide/en/elasticsearch/reference/master/sort-search-results.html

        Returns:
            A dataframe containing all matches as rows.
        """
        query_args = (field, value, regexp)
        all_hits = self.all_matches_for_query(self.get_n_datasets_for_field,
                                              query_args,
                                              sort_by)
        search_df = json_normalize(all_hits)
        return search_df

    def get_n_datasets_for_field_combination(
            self,
            and_fields,
            or_fields,
            not_fields,
            n=Discover.DEFAULT_PAGE_SIZE,
            cursor=None,
            sort_by=Discover.DEFAULT_SORT_ORDER):
        """Fetch dataset entries that satisfy the given match combinations.

        Args:
            and_fields (dict):
                A dictionary of fields as keys mapping to their expected
                values. All fields within this dictionary must match.
            or_fields (dict):
                A dictionary of fields as keys mapping to their expected
                values. Any of the fields can be present in the match.
            not_fields (dict):
                A dictionary of fields as keys mapping to their expected
                values. None of the fields must match.
            n (int, optional):
                Number of matches to fetch. Defaults to
                `Discover.DEFAULT_PAGE_SIZE`.
            cursor (list, optional):
                A cursor from where to begin searching. Can be obtained from
                calls to this method. Defaults to None, in case search needs to
                start at the beginning.
            sort_by (list):
                A list of dictionaries which will be used to order responses
                for queries. Defaults to sorting based on combination of score
                (descending) and ID (ascending).

                Each dictionary is a key-value pair, where the keys are fields
                to be sorted on and the values may contain order, mode, etc. as
                documented here:
                https://www.elastic.co/guide/en/elasticsearch/reference/master/sort-search-results.html

        Returns:
            A list of hits and a cursor. The list contains the `n` hits as
            rows. The cursor can be used to fetch the next set of items from
            where this one ended.
        """
        and_match_list = [{"match": {k: v}} for k, v in and_fields.items()]
        or_match_list = [{"match": {k: v}} for k, v in or_fields.items()]
        not_match_list = [{"match": {k: v}} for k, v in not_fields.items()]

        payload = {
            "query": {
                "bool": {
                    "must": and_match_list,
                    "must_not": not_match_list,
                    "should": or_match_list
                }
            },
        }
        return self._execute_search_for_hits(payload, n, cursor, sort_by)

    def query_dataset_by_field_combination(self,
                                           and_fields={},
                                           or_fields={},
                                           not_fields={},
                                           n=Discover.DEFAULT_PAGE_SIZE):
        print("Warning: this method has been deprecated. Please consider",
              "using 'get_n_datasets_for_field_combination_as_df' instead.")

        hits, _ = self.get_n_datasets_for_field_combination(and_fields,
                                                            or_fields,
                                                            not_fields,
                                                            n,
                                                            None)
        search_df = json_normalize(hits)
        return search_df

    def get_n_datasets_for_field_combination_as_df(
            self,
            and_fields={},
            or_fields={},
            not_fields={},
            n=Discover.DEFAULT_PAGE_SIZE,
            cursor=None,
            sort_by=Discover.DEFAULT_SORT_ORDER):
        """Obtain a dataframe of specified number of dataset matches that
        satisfy the given match combinations.

        Args:
            and_fields (dict):
                A dictionary of fields as keys mapping to their expected
                values. All fields within this dictionary must match.
            or_fields (dict):
                A dictionary of fields as keys mapping to their expected
                values. Any of the fields can be present in the match.
            not_fields (dict):
                A dictionary of fields as keys mapping to their expected
                values. None of the fields must match.
            n (int, optional):
                Number of matches to fetch. Defaults to
                `Discover.DEFAULT_PAGE_SIZE`.
            cursor (list, optional):
                A cursor from where to begin searching. Can be obtained from
                calls to this method. Defaults to None, in case search needs to
                start at the beginning.
            sort_by (list):
                A list of dictionaries which will be used to order responses
                for queries. Defaults to sorting based on combination of score
                (descending) and ID (ascending).

                Each dictionary is a key-value pair, where the keys are fields
                to be sorted on and the values may contain order, mode, etc. as
                documented here:
                https://www.elastic.co/guide/en/elasticsearch/reference/master/sort-search-results.html

        Returns:
            A dataframe and a cursor. The dataframe contains the `n` hits as
            rows. The cursor can be used to fetch the next set of items from
            where this one ended.
        """
        hits, next_cursor = self.get_n_datasets_for_field_combination(
            and_fields, or_fields, not_fields, n, cursor, sort_by)
        search_df = json_normalize(hits)
        return search_df, next_cursor

    def get_all_datasets_for_field_combination_as_df(
            self,
            and_fields={},
            or_fields={},
            not_fields={},
            sort_by=Discover.DEFAULT_SORT_ORDER):
        """Obtain a dataframe of all datasets that satisfy the given match
        combinations.

        Args:
            and_fields (dict):
                A dictionary of fields as keys mapping to their expected
                values. All fields within this dictionary must match.
            or_fields (dict):
                A dictionary of fields as keys mapping to their expected
                values. Any of the fields can be present in the match.
            not_fields (dict):
                A dictionary of fields as keys mapping to their expected
                values. None of the fields must match.
            sort_by (list):
                A list of dictionaries which will be used to order responses
                for queries. Defaults to sorting based on combination of score
                (descending) and ID (ascending).

                Each dictionary is a key-value pair, where the keys are fields
                to be sorted on and the values may contain order, mode, etc. as
                documented here:
                https://www.elastic.co/guide/en/elasticsearch/reference/master/sort-search-results.html

        Returns:
            A dataframe containing all matches as rows.
        """
        query_args = (and_fields, or_fields, not_fields)
        all_hits = self.all_matches_for_query(
            self.get_n_datasets_for_field_combination, query_args, sort_by)
        search_df = json_normalize(all_hits)
        return search_df

    def query_dataset_by_field_autocomplete(self, field, value, autofield, regexp = False, n = 10000):
        payload = {
        "_source": ["organism",autofield],
          "query": {
              "match": {field: value}
          },
        'size' : n
        }
        url =  self.host + '/' + self.endpoint + '/_search'
        r = make_discover_request('GET', url, json=payload)
        if r.status_code != 200:
            r = make_discover_request('GET', url, json=payload)
        if r.status_code != 200:
            r = make_discover_request('GET', url, json=payload)
        if r.status_code != 200:
            return r.status_code
        output = r.json()
        search_df = json_normalize(output["hits"]["hits"])
        return search_df
