import requests
import json
import pandas as pd
from pandas import json_normalize

from .env import (ELASTIC_DOMAIN_URL, AWS_REGION,
    DISCOVER_API_URL)
from .helpers import make_discover_request
from . import constants as cnst
from .constants import (BULK_FILES_INDEX, BULK_SAMPLE_REPO_INDEX,
                        BULK_DATA_INDEX, SC_FILES_INDEX,
                        SC_SAMPLE_REPO_INDEX, SC_DATA_INDEX,
                        REPOSITORY_PACKAGE_ENDPOINT)
from .download_helper import download_file


class Discover:
    DEFAULT_PAGE_SIZE = 100
    DEFAULT_SORT_ORDER = [{"_score": "desc"}, {"_id": "asc"}]

    def __init__(self, host = ELASTIC_DOMAIN_URL, region=AWS_REGION):
        """
        returns a discover object
        """
        self.host = host
        self.service = 'es'
        self.region = region
        self.endpoint = None

    def get_repositories(self):
        url = DISCOVER_API_URL+cnst.REPOSITORIES_ENDPOINT
        resp = make_discover_request('GET', url)
        if resp.status_code == cnst.OK:
            data = resp.json().get('data')
            if data:
                repositories = [each.get('attributes') for each in data]
                return repositories
        else:
            print(resp.text)

    def set_repo(self, repo_id, mode = 'bulk'):

        if not isinstance(repo_id, str):
            raise Exception('repo_id must be a str')

        if mode not in ["bulk", "single_cell"]:
            raise Exception("Mode has to be one of 'bulk' or 'single_cell'")

        repositories = self.get_repositories()
        repo = None
        for each in repositories:
            if each['repo_id'] == repo_id:
                repo = each
                break

        if repo and repo.get('indexes'):
            indexes = repo.get('indexes')
            if mode == "bulk":
                files_endpoint = indexes.get(BULK_FILES_INDEX)
                sample_repo_endpoint = indexes.get(BULK_SAMPLE_REPO_INDEX)
                data_endpoint = indexes.get(BULK_DATA_INDEX)

            if mode == "single_cell":
                files_endpoint = indexes.get(SC_FILES_INDEX)
                sample_repo_endpoint = indexes.get(SC_SAMPLE_REPO_INDEX)
                data_endpoint = indexes.get(SC_DATA_INDEX)

            self.set_sample_repo(sample_repo_endpoint)
            self.set_dataset_repo(files_endpoint)
            self.set_feature_repo(data_endpoint)
            self.package = self.get_repo_package(repo_id)

        else:
            print('Error while setting the repo. Try Discover().get_repositories()')


    def set_sample_repo(self, endpoint):
        from .sample_repo import SampleRepo
        self.sample_repo = SampleRepo(endpoint)
        self.sample_repo.host = self.host
        self.sample_repo.service = self.service
        self.sample_repo.region = self.region

    def set_feature_repo(self, endpoint):
        from .feature_repo import FeatureRepo
        self.feature_repo = FeatureRepo(endpoint)
        self.feature_repo.host = self.host
        self.feature_repo.service = self.service
        self.feature_repo.region = self.region

    def set_dataset_repo(self, endpoint):
        from .dataset_repo import DatasetRepo
        self.dataset_repo = DatasetRepo(endpoint)
        self.dataset_repo.host = self.host
        self.dataset_repo.service = self.service
        self.dataset_repo.region = self.region
        # hotfix
        endpoint = endpoint.split("_")[0]+"_csv"
        from .annotation_repo import AnnotationRepo
        self.annotation_repo = AnnotationRepo(endpoint)
        self.annotation_repo.host = self.host
        self.annotation_repo.service = self.service
        self.annotation_repo.region = self.region

    def set_annotation_repo(self, endpoint):
        from .annotation_repo import AnnotationRepo
        self.annotation_repo = AnnotationRepo(endpoint)
        self.annotation_repo.host = self.host
        self.annotation_repo.service = self.service
        self.annotation_repo.region = self.region


    def __repr__(self):
        s = "Discover object"
        s = s + "\n" + f'host: {self.host}'
        s = s + "\n" + f'region: {self.region}'
        if hasattr(self, "dataset_repo"):
            s = s + "\n" + f'dataset_repo: {self.dataset_repo.endpoint}'

        if hasattr(self, "sample_repo"):
            s = s+ "\n" + f'sample_repo: {self.sample_repo.endpoint}'

        if hasattr(self, "feature_repo"):
            s = s + "\n"+ f'feature_repo: {self.feature_repo.endpoint}'

        if hasattr(self, "annotation_repo"):
            s = s + "\n" + f'annotation_repo: {self.annotation_repo.endpoint}'

        return s

    def get_all_fields(self):
        """
        get all fields available for querying the endpoint
        """
        url = self.host + '/' + self.endpoint
        print(url)
        r = make_discover_request('GET', url)
        output = r.json()
        fields = list(output[self.endpoint]["mappings"]["properties"].keys())
        return fields

    def get_n_records(self, n=10, cursor=None):
        """Fetch the top `n` records from a repository.

        If cursor is not `None` then, it should be a list whose first element
        is used as an offset after which `n` records are pulled.

        Args:
            n (int, optional):
                The number of records to fetch.
            cursor (list, optional):
                A list whose first element used as an offset. Defaults to None.

        Returns:
            A pair of lists. The first list is the `n` records that were
            fetched. The second is a single-element list meant to be used as a
            cursor for the next set of records.
        """
        url = self.host + '/' + self.endpoint + '/_search'
        payload = {"size": n, "sort": {"_id": "asc"}}

        if cursor is not None:
            payload["search_after"] = cursor

        r = make_discover_request('GET', url, params=payload)
        output = r.json()

        hits = []
        if "hits" in output:
            hits = output["hits"]["hits"]
        records = [each["_source"] for each in list(hits)]

        next_cursor = []
        if len(hits) > 0:
            last_element = hits[-1]
            next_cursor = [last_element["_id"]]

        return records, next_cursor

    def get_top_n_examples(self, n=10):
        """Obtain a dataframe of the top `n` records in this repository.

        Args:
            n (int, optional):
                The number of records to fetch.

        Returns:
            A dataframe contains the `n` records as rows.
        """
        print("Warning: this method has been deprecated.",
              "Please consider using 'get_n_records_as_df' instead.")
        records, _ = self.get_n_records(n, None)
        top_n_df = json_normalize(records)
        return top_n_df

    def get_n_records_as_df(self, n=10, cursor=None):
        """Obtain a dataframe of the top `n` records in this repository.

        Args:
            n (int, optional):
                The number of records to fetch.
            cursor (list, optional):
                A list whose first element used as an offset. Defaults to None.

        Returns:
            A dataframe and a cursor. The dataframe contains the `n` records as
            rows. The cursor can be used to fetch the next set of items from
            where this one ended.
        """
        records, next_cursor = self.get_n_records(n, cursor)
        top_n_df = json_normalize(records)
        return top_n_df, next_cursor

    def get_all_records_as_df(self):
        """Obtain a dataframe of all records in this repository.

        Returns:
            A dataframe containing the records as rows.
        """
        all_records = self.all_matches_for_query(self.get_n_records, tuple())
        df = json_normalize(all_records)
        return df

    def all_matches_for_query(self, query_fn, query_args, sort_by):
        """Given a function and its initial arguments, iteratively call it
        until it returns an empty cursor. After each iteration the results are
        concatenated with previous results.

        Args:
            query_fn (function):
                A function that takes a set of arguments - the last of which
                should be the number of items to be fetched and a cursor (list)
                object. The function should return a list of items fetched and
                a cursor to be passed to the next call for items.
            query_args (tuple):
                Arguments for `query_fn` excluding the last two (size and
                cursor parameters).
            sort_by (list):
                A list of dictionaries which will be used to order responses
                for queries. Each dictionary is a key-value pair, where the
                keys are fields to be sorted on and the values may contain
                order, mode, etc. as documented here:
                https://www.elastic.co/guide/en/elasticsearch/reference/master/sort-search-results.html

        Returns:
            A combined list of all items fetched from the iterative calls to
            `query_fn`.
        """
        size = Discover.DEFAULT_PAGE_SIZE
        all_items = []

        items, cursor = query_fn(*query_args, size, None, sort_by)
        while len(cursor) > 0:
            all_items += items
            items, cursor = query_fn(*query_args, size, cursor, sort_by)

        return all_items

    def _execute_search_for_hits(self, payload, n, cursor, sort_by):
        """For a given cursor, execute search beginning at that cursor for the
        given payload.

        Args:
            payload (dict):
                A valid payload for performing an elasticsearch query.
            n (int):
                Number of matches to fetch.
            cursor (list):
                A cursor to tell from where to begin searching.
            sort_by (list):
                A list of dictionaries which will be used to order responses
                for queries.

        Returns:
            A list of hits and a cursor.
        """
        if not isinstance(payload, dict):
            payload = {}

        payload["size"] = n
        payload["sort"] = sort_by
        if cursor is not None:
            payload["search_after"] = cursor

        url = self.host + '/' + self.endpoint + '/_search'
        r = make_discover_request('GET', url, json=payload)
        output = r.json()

        hits = []
        if "hits" in output:
            hits = output["hits"]["hits"]

        next_cursor = []
        if len(hits) > 0:
            last_element = hits[-1]
            next_cursor = [last_element["_score"], last_element["_id"]]

        return hits, next_cursor

    def get_file(self, key, repo_id, local_fp):
        download_file(key, repo_id, local_fp)

    def get_repo_package(self, repo_id):
        url = DISCOVER_API_URL+ str.format(REPOSITORY_PACKAGE_ENDPOINT, repo_id)
        resp = make_discover_request('GET', url)
        if resp.status_code == cnst.OK:
            data = resp.json().get('data')
            assert len(data) < 2 , "Multiple packages exist for this repo id" 
            if len(data)==1:
                package_name = data[0].get('attributes').get('package_name')
                return package_name
            return ""
        else:
            print(resp.text)