import pandas as pd

from .discover import Discover
from .helpers import get_dataset_name_from_doc_id


class FeatureRepo(Discover):
    def __init__(self, endpoint):
        self.endpoint = endpoint

    def get_n_feature_match_values(self,
                                   feature_name,
                                   n=Discover.DEFAULT_PAGE_SIZE,
                                   cursor=None,
                                   sort_by=Discover.DEFAULT_SORT_ORDER):
        """Fetch feature values matching against the `feature_name` term. The
        `n` here decides the number of datasets which will be searched.
        Multiple matches per dataset may result in match counts that are
        greater than `n`.

        Args:
            feature_name (str):
                Name of the feature being searched for.
            n (int, optional):
                Number of datasets from which to fetch matches from.
            cursor (list, optional):
                A cursor from where to begin searching. Can be obtained from
                calls to this method. Defaults to None, in case search needs to
                start at the beginning.
            sort_by (list):
                A list of dictionaries which will be used to order responses
                for queries. Defaults to sorting based on combination of score
                (descending) and ID (ascending).

                Each dictionary is a key-value pair, where the keys are fields
                to be sorted on and the values may contain order, mode, etc. as
                documented here:
                https://www.elastic.co/guide/en/elasticsearch/reference/master/sort-search-results.html

        Returns:
            A list of hits and a cursor. The list contains all feature value
            matches, for at most `n` datasets. The cursor can be used to fetch
            the next set of items from where this one ended.
        """
        payload = {
            "query": {
                "match": {
                    "kw_index": feature_name
                }
            },
        }

        hits, next_cursor = self._execute_search_for_hits(payload,
                                                          n,
                                                          cursor,
                                                          sort_by)

        feature_sample_values = []
        for doc_dict in hits:
            dataset_name = get_dataset_name_from_doc_id(
                doc_dict["_source"]["kw_doc_id"])
            feature_name = doc_dict["_source"]["kw_index"]
            sample_values = [
                [feature_name] + list(d.values()) + [dataset_name]
                for d in doc_dict['_source']['kw_column']
                if 'kw_index' not in d.values()
            ]
            feature_sample_values += sample_values

        return feature_sample_values, next_cursor

    def get_feature_values(self, feature_name, n=Discover.DEFAULT_PAGE_SIZE):
        print("Warning: this method has been deprecated. Please consider",
              "using 'get_n_feature_match_values_as_df' instead.")

        values, _ = self.get_n_feature_match_values(feature_name, n, None)
        feature_sample_df = pd.DataFrame(values)
        if len(feature_sample_df) > 0:
            feature_sample_df.columns = [
                "feature_name", "sample", "value", "dataset"
            ]
        return feature_sample_df

    def get_n_feature_match_values_as_df(self,
                                         feature_name,
                                         n=Discover.DEFAULT_PAGE_SIZE,
                                         cursor=None,
                                         sort_by=Discover.DEFAULT_SORT_ORDER):
        """Obtain a dataframe of feature values matching against the
        `feature_name` term. The `n` here decides the number of datasets which
        will be searched. Multiple matches per dataset may result in match
        counts that are greater than `n`.

        Args:
            feature_name (str):
                Name of the feature being searched for.
            n (int, optional):
                Number of datasets from which to fetch matches from.
            cursor (list, optional):
                A cursor from where to begin searching. Can be obtained from
                calls to this method. Defaults to None, in case search needs to
                start at the beginning.
            sort_by (list):
                A list of dictionaries which will be used to order responses
                for queries. Defaults to sorting based on combination of score
                (descending) and ID (ascending).

                Each dictionary is a key-value pair, where the keys are fields
                to be sorted on and the values may contain order, mode, etc. as
                documented here:
                https://www.elastic.co/guide/en/elasticsearch/reference/master/sort-search-results.html

        Returns:
            A dataframe and a cursor. The dataframe contains all feature value
            matches, for at most `n` datasets, as rows. The cursor can be used
            to fetch the next set of items from where this one ended.
        """
        values, next_cursor = self.get_n_feature_match_values(feature_name,
                                                              n,
                                                              cursor,
                                                              sort_by)
        feature_sample_df = pd.DataFrame(values)
        if len(feature_sample_df) > 0:
            feature_sample_df.columns = [
                "feature_name", "sample", "value", "dataset"
            ]
        return feature_sample_df, next_cursor

    def get_all_feature_match_values_as_df(
            self,
            feature_name,
            sort_by=Discover.DEFAULT_SORT_ORDER):
        """Obtain a dataframe of feature values matching against the
        `feature_name` term, across all datasets.

        Args:
            feature_name (str):
                Name of the feature being searched for.
            sort_by (list):
                A list of dictionaries which will be used to order responses
                for queries. Defaults to sorting based on combination of score
                (descending) and ID (ascending).

                Each dictionary is a key-value pair, where the keys are fields
                to be sorted on and the values may contain order, mode, etc. as
                documented here:
                https://www.elastic.co/guide/en/elasticsearch/reference/master/sort-search-results.html

        Returns:
            A dataframe containing feature value matches, across all datasets.
        """
        query_args = (feature_name,)
        all_values = self.all_matches_for_query(
            self.get_n_feature_match_values, query_args, sort_by)
        feature_sample_df = pd.DataFrame(all_values)
        if len(feature_sample_df) > 0:
            feature_sample_df.columns = [
                "feature_name", "sample", "value", "dataset"
            ]
        return feature_sample_df

    def get_n_sc_feature_match_values(self,
                                      feature_name,
                                      n=Discover.DEFAULT_PAGE_SIZE,
                                      cursor=None,
                                      sort_by=Discover.DEFAULT_SORT_ORDER):
        """Fetch feature values matching against the `feature_name` term, in
        single-cell datasets. The `n` here decides the number of datasets which
        will be searched. Multiple matches per dataset may result in match
        counts that are greater than `n`.

        Args:
            feature_name (str):
                Name of the feature being searched for.
            n (int, optional):
                Number of SC datasets from which to fetch matches from.
            cursor (list, optional):
                A cursor from where to begin searching. Can be obtained from
                calls to this method. Defaults to None, in case search needs to
                start at the beginning.
            sort_by (list):
                A list of dictionaries which will be used to order responses
                for queries. Defaults to sorting based on combination of score
                (descending) and ID (ascending).

                Each dictionary is a key-value pair, where the keys are fields
                to be sorted on and the values may contain order, mode, etc. as
                documented here:
                https://www.elastic.co/guide/en/elasticsearch/reference/master/sort-search-results.html

        Returns:
            A list of hits and a cursor. The list contains all feature value
            matches, for at most `n` datasets. The cursor can be used to fetch
            the next set of items from where this one ended.
        """
        payload = {
            "query": {
                "match": {
                    "kw_index": feature_name
                }
            },
        }

        hits, next_cursor = self._execute_search_for_hits(payload,
                                                          n,
                                                          cursor,
                                                          sort_by)

        feature_sample_values = []
        for doc_dict in hits:
            dataset_name = doc_dict["_source"]["kw_doc_id"]
            feature_name = doc_dict["_source"]["kw_index"]
            sample_values = [
                [feature_name] + [k] + [v] + [dataset_name]
                for k, v in doc_dict['_source'].items()
                if "geneexp" in k
            ]
            feature_sample_values += sample_values

        return feature_sample_values, next_cursor

    def get_feature_values_sc(self,
                              feature_name,
                              n=Discover.DEFAULT_PAGE_SIZE):
        print("Warning: this method has been deprecated. Please consider",
              "using 'get_n_sc_feature_matches_as_df' instead.")

        feature_sample_values, _ = self.get_n_sc_feature_match_values(
            feature_name, n, None)
        feature_sample_df = pd.DataFrame(feature_sample_values)
        if len(feature_sample_df) > 0:
            feature_sample_df.columns = [
                "feature_name", "celltype", "value", "dataset"
            ]
        return feature_sample_df

    def get_n_sc_feature_match_values_as_df(
            self,
            feature_name,
            n=Discover.DEFAULT_PAGE_SIZE,
            cursor=None,
            sort_by=Discover.DEFAULT_SORT_ORDER):
        """Obtain a dataframe of feature values matching against the
        `feature_name` term, in single-cell datasets. The `n` here decides the
        number of datasets which will be searched. Multiple matches per dataset
        may result in match counts that are greater than `n`.

        Args:
            feature_name (str):
                Name of the feature being searched for.
            n (int, optional):
                Number of SC datasets from which to fetch matches from.
            cursor (list, optional):
                A cursor from where to begin searching. Can be obtained from
                calls to this method. Defaults to None, in case search needs to
                start at the beginning.
            sort_by (list):
                A list of dictionaries which will be used to order responses
                for queries. Defaults to sorting based on combination of score
                (descending) and ID (ascending).

                Each dictionary is a key-value pair, where the keys are fields
                to be sorted on and the values may contain order, mode, etc. as
                documented here:
                https://www.elastic.co/guide/en/elasticsearch/reference/master/sort-search-results.html

        Returns:
            A dataframe and a cursor. The dataframe contains all feature value
            matches, for at most `n` datasets, as rows. The cursor can be used
            to fetch the next set of items from where this one ended.
        """
        feature_sample_values, next_cursor = \
            self.get_n_sc_feature_match_values(feature_name,
                                               n,
                                               cursor,
                                               sort_by)
        feature_sample_df = pd.DataFrame(feature_sample_values)
        if len(feature_sample_df) > 0:
            feature_sample_df.columns = [
                "feature_name", "celltype", "value", "dataset"
            ]
        return feature_sample_df, next_cursor

    def get_all_sc_feature_match_values_as_df(
            self,
            feature_name,
            sort_by=Discover.DEFAULT_SORT_ORDER):
        """Obtain a dataframe of feature values matching against the
        `feature_name` term, across all single-cell datasets.

        Args:
            feature_name (str):
                Name of the feature being searched for.
            sort_by (list):
                A list of dictionaries which will be used to order responses
                for queries. Defaults to sorting based on combination of score
                (descending) and ID (ascending).

                Each dictionary is a key-value pair, where the keys are fields
                to be sorted on and the values may contain order, mode, etc. as
                documented here:
                https://www.elastic.co/guide/en/elasticsearch/reference/master/sort-search-results.html

        Returns:
            A dataframe containing feature value matches, across all
            single-cell datasets.
        """
        query_args = (feature_name,)
        all_values = self.all_matches_for_query(
            self.get_n_sc_feature_match_values, query_args, sort_by)
        feature_sample_df = pd.DataFrame(all_values)
        if len(feature_sample_df) > 0:
            feature_sample_df.columns = [
                "feature_name", "sample", "value", "dataset"
            ]
        return feature_sample_df
