from . import constants as const
from .env import DISCOVER_API_URL
from .helpers import make_discover_request


class Package(object):
    """
    Package Model
    """

    REPOSITORY_PACKAGE_URL = DISCOVER_API_URL + const.REPOSITORY_PACKAGE_ENDPOINT

    def __init__(self, package_name, repo_id):
        """
        Create a new Package object

        Args:
            package_name (str): Package to be created
            repo_id (str): Repository ID

        Raises:
            ValueError: if package_name is not a string or is empty
            ValueError: if repo_id is not a string/int or is empty
        """
        if not isinstance(package_name, str) or not package_name:
            raise ValueError("package_name must be a str and cannot be empty")

        if (
            not isinstance(repo_id, str) and not isinstance(repo_id, int)
        ) or not repo_id:
            raise ValueError("repo_id must be a str or int and cannot be empty")

        self.package_name = package_name
        self.repo_id = str(repo_id)

        self.REPOSITORY_PACKAGE_URL = self.REPOSITORY_PACKAGE_URL.format(repo_id)

    def create(self):
        """
        Create an entry in Package table

        Raises:
            Exception: if Package already exists in the repository
            Exception: if Package creation fails

        Returns:
            package (Package): created package response
        """
        repository_packages = self.get(self.repo_id)
        if len(repository_packages) > 0:
            raise Exception("package already exists for this repository")

        payload = const.CREATE_PACKAGE_PAYLOAD.copy()
        payload["data"]["attributes"]["package_name"] = self.package_name
        resp = make_discover_request("POST", self.REPOSITORY_PACKAGE_URL, json=payload)
        if resp.status_code == const.CREATED:
            created_package = resp.json().get("data")
        else:
            raise Exception(resp.text)

        return created_package

    @classmethod
    def get(cls, repo_id):
        """
        Get packages of a repository

        Args:
            repo_id (str): Repository ID

        Raises:
            Exception: if failed to get packages

        Returns:
            packages (list[packages]): Packages
        """
        if (
            not isinstance(repo_id, str) and not isinstance(repo_id, int)
        ) or not repo_id:
            raise ValueError("repo_id must be a str or int and cannot be empty")

        resp = make_discover_request(
            "GET", cls.REPOSITORY_PACKAGE_URL.format(str(repo_id))
        )
        if resp.status_code == const.OK:
            packages = resp.json().get("data")
        else:
            raise Exception(resp.text)

        return packages
