from . import constants as const
from .env import DISCOVER_API_URL
from .helpers import make_discover_request
from .package import Package


class Repository(object):
    """
    Repository Model
    """

    REPOSITORY_URL = DISCOVER_API_URL + const.REPOSITORIES_ENDPOINT

    def __init__(
        self,
        display_name,
        description,
        initials,
        explorer_enabled=False,
        repo_name=None,
    ):
        """
        Create a repository object

        Args:
            display_name (str): Repository display_name
            description (str): Repository description
            initials (str): Repository Initials
            explorer_enabled (bool, optional): Whether Explorer is enabled. Defaults to False.
            repo_name (str, optional): Internal repository name. Defaults to display_name.lower().replace(" ", "_").

        Raises:
            ValueError: if display_name is not a string or empty
            ValueError: if description is not a string or empty
            ValueError: if initials is not a string or empty
        """
        if not isinstance(display_name, str) or not display_name:
            raise ValueError(f"display_name must be str and cannot be empty")

        if not isinstance(description, str) or not description:
            raise ValueError(f"description must be str and cannot be empty")

        if not isinstance(initials, str) or not initials:
            raise ValueError(f"initials must be str and cannot be empty")

        self.display_name = display_name
        self.description = description
        self.initials = initials
        self.explorer_enabled = explorer_enabled

        if repo_name is None:
            self.repo_name = self.display_name.lower().replace(" ", "_")
        else:
            self.repo_name = repo_name

    def create(self):
        """
        Create an entry in Repository table

        Raises:
            Exception: if Repository creation failed

        Returns:
            repository (Repository): created repository
        """
        payload = const.CREATE_REPOSITORY_PAYLOAD.copy()
        payload["data"]["attributes"]["repo_name"] = self.repo_name

        frontend_info = payload["data"]["attributes"]["frontend_info"]
        frontend_info["display_name"] = self.display_name
        frontend_info["description"] = self.description
        frontend_info["initials"] = self.initials
        frontend_info["explorer_enabled"] = self.explorer_enabled

        indexes = payload["data"]["attributes"]["indexes"]
        for key in indexes.keys():
            indexes[key] = f"{self.repo_name}_{key}"

        resp = make_discover_request("POST", self.REPOSITORY_URL, json=payload)
        if resp.status_code == const.CREATED:
            created_repository = resp.json().get("data")
        else:
            raise Exception(resp.text)

        self.repo_id = created_repository["id"]
        return created_repository

    def add_package(self, package_name):
        """
        Create a new package in the repository

        Args:
            package_name (str): Package to be created

        Returns:
            package (Package): created package
        """
        package = Package(package_name, self.repo_id)
        created_package = package.create()
        return created_package
