from pandas import json_normalize

from .discover import Discover
from .helpers import generate_dualcase_string


class SampleRepo(Discover):
    def __init__(self, endpoint):
        self.endpoint = endpoint

    def get_n_samples_for_field(self,
                                field,
                                value,
                                regexp,
                                n=Discover.DEFAULT_PAGE_SIZE,
                                cursor=None,
                                sort_by=Discover.DEFAULT_SORT_ORDER):
        """Fetch sample entries that match against the `value` for a `field`.
        Optionally a regular expression (`regexp`) can be provided instead of a
        `value`.

        Args:
            field (str):
                Name of the field whose values are to be matched against.
            value (str):
                Expression to be searched for. Must be a valid Elasticsearch
                "query" string.
            regexp (str):
                A regular expression in case more flexibility is desired rather
                than a plain `value`.
            n (int, optional):
                Number of sample matches to fetch. Defaults to
                `Discover.DEFAULT_PAGE_SIZE`.
            cursor (list, optional):
                A cursor from where to begin searching. Can be obtained from
                calls to this method. Defaults to None, in case search needs to
                start at the beginning.
            sort_by (list):
                A list of dictionaries which will be used to order responses
                for queries. Defaults to sorting based on combination of score
                (descending) and ID (ascending).

                Each dictionary is a key-value pair, where the keys are fields
                to be sorted on and the values may contain order, mode, etc. as
                documented here:
                https://www.elastic.co/guide/en/elasticsearch/reference/master/sort-search-results.html

        Returns:
            A list of hits and a cursor. The list contains the `n` hits as
            rows. The cursor can be used to fetch the next set of items from
            where this one ended.
        """
        payload = {"query": {"match": {field: value}}}
        if regexp:
            payload = {
                "query": {
                    "regexp": {
                        f"{field}.keyword": {
                            "value": f".*{generate_dualcase_string(value)}.*",
                            "flags": "ALL",
                            "max_determinized_states": 10000,
                            "rewrite": "constant_score"
                        }
                    }
                },
            }

        return self._execute_search_for_hits(payload, n, cursor, sort_by)

    def query_samples_by_field(self,
                               field,
                               value,
                               regexp=False,
                               n=Discover.DEFAULT_PAGE_SIZE):
        print("Warning: this method has been deprecated. Please consider",
              "using 'get_n_samples_for_field_as_df' instead.")

        hits, _ = self.get_n_samples_for_field(field, value, regexp, n, None)
        search_df = json_normalize(hits)
        return search_df

    def get_n_samples_for_field_as_df(self,
                                      field,
                                      value,
                                      regexp=False,
                                      n=Discover.DEFAULT_PAGE_SIZE,
                                      cursor=None,
                                      sort_by=Discover.DEFAULT_SORT_ORDER):
        """Obtain a dataframe of specified number of sample matches against
        the `value` for a `field`. Optionally a regular expression (`regexp`)
        can be provided instead of a `value`.

        Args:
            field (str):
                Name of the field whose values are to be matched against.
            value (str):
                Expression to be searched for. Must be a valid Elasticsearch
                "query" string.
            regexp (str):
                A regular expression in case more flexibility is desired rather
                than a plain `value`.
            n (int, optional):
                Number of matches to fetch. Defaults to
                `Discover.DEFAULT_PAGE_SIZE`.
            cursor (list, optional):
                A cursor from where to begin searching. Can be obtained from
                calls to this method. Defaults to None, in case search needs to
                start at the beginning.
            sort_by (list):
                A list of dictionaries which will be used to order responses
                for queries. Defaults to sorting based on combination of score
                (descending) and ID (ascending).

                Each dictionary is a key-value pair, where the keys are fields
                to be sorted on and the values may contain order, mode, etc. as
                documented here:
                https://www.elastic.co/guide/en/elasticsearch/reference/master/sort-search-results.html

        Returns:
            A dataframe and a cursor. The dataframe contains the `n` hits as
            rows. The cursor can be used to fetch the next set of items from
            where this one ended.
        """
        hits, next_cursor = self.get_n_samples_for_field(field,
                                                         value,
                                                         regexp,
                                                         n,
                                                         cursor,
                                                         sort_by)
        search_df = json_normalize(hits)
        return search_df, next_cursor

    def get_all_samples_for_field_as_df(self,
                                        field,
                                        value,
                                        regexp=False,
                                        sort_by=Discover.DEFAULT_SORT_ORDER):
        """Obtain a dataframe of all sample matches against the `value` for
        a `field`. Optionally a regular expression (`regexp`) can be provided
        instead of a `value`.

        Args:
            field (str):
                Name of the field whose values are to be matched against.
            value (str):
                Expression to be searched for. Must be a valid Elasticsearch
                "query" string.
            regexp (str):
                A regular expression in case more flexibility is desired rather
                than a plain `value`.
            sort_by (list):
                A list of dictionaries which will be used to order responses
                for queries. Defaults to sorting based on combination of score
                (descending) and ID (ascending).

                Each dictionary is a key-value pair, where the keys are fields
                to be sorted on and the values may contain order, mode, etc. as
                documented here:
                https://www.elastic.co/guide/en/elasticsearch/reference/master/sort-search-results.html

        Returns:
            A dataframe containing all matches as rows.
        """
        query_args = (field, value, regexp)
        all_hits = self.all_matches_for_query(self.get_n_samples_for_field,
                                              query_args,
                                              sort_by)
        search_df = json_normalize(all_hits)
        return search_df

    def get_n_samples_for_fields(self,
                                 search_term,
                                 fields=[],
                                 n=Discover.DEFAULT_PAGE_SIZE,
                                 cursor=None,
                                 sort_by=Discover.DEFAULT_SORT_ORDER):
        """Fetch sample entries that match against the `search_term` in any of
        the given `fields`.

        Args:
            search_term (str):
                Expression to be searched for. Must be a valid Elasticsearch
                "query" string.
            fields (list):
                Name of the fields whose values are to be matched against.
            n (int, optional):
                Number of sample matches to fetch. Defaults to
                `Discover.DEFAULT_PAGE_SIZE`.
            cursor (list, optional):
                A cursor from where to begin searching. Can be obtained from
                calls to this method. Defaults to None, in case search needs to
                start at the beginning.
            sort_by (list):
                A list of dictionaries which will be used to order responses
                for queries. Defaults to sorting based on combination of score
                (descending) and ID (ascending).

                Each dictionary is a key-value pair, where the keys are fields
                to be sorted on and the values may contain order, mode, etc. as
                documented here:
                https://www.elastic.co/guide/en/elasticsearch/reference/master/sort-search-results.html

        Returns:
            A list of hits and a cursor. The list contains the `n` hits as
            rows. The cursor can be used to fetch the next set of items from
            where this one ended.
        """
        payload = {
            "query": {
                "multi_match": {
                    "query": search_term,
                    "fields": fields
                }
            },
        }
        return self._execute_search_for_hits(payload, n, cursor, sort_by)

    def query_samples_by_all_fields(self,
                                    search_term,
                                    fields=[],
                                    n=Discover.DEFAULT_PAGE_SIZE):
        print("Warning: this method has been deprecated. Please consider",
              "using 'get_n_samples_for_fields_as_df' instead.")

        hits, _ = self.get_n_samples_for_fields(search_term, fields, n, None)
        search_df = json_normalize(hits)
        return search_df

    def get_n_samples_for_fields_as_df(self,
                                       search_term,
                                       fields=[],
                                       n=Discover.DEFAULT_PAGE_SIZE,
                                       cursor=None,
                                       sort_by=Discover.DEFAULT_SORT_ORDER):
        """Obtain a dataframe of specified number of sample matches against the
        `search_term` in any of the given `fields`.

        Args:
            search_term (str):
                Expression to be searched for. Must be a valid Elasticsearch
                "query" string.
            fields (list):
                Name of the fields whose values are to be matched against.
            n (int, optional):
                Number of sample matches to fetch. Defaults to
                `Discover.DEFAULT_PAGE_SIZE`.
            cursor (list, optional):
                A cursor from where to begin searching. Can be obtained from
                calls to this method. Defaults to None, in case search needs to
                start at the beginning.
            sort_by (list):
                A list of dictionaries which will be used to order responses
                for queries. Defaults to sorting based on combination of score
                (descending) and ID (ascending).

                Each dictionary is a key-value pair, where the keys are fields
                to be sorted on and the values may contain order, mode, etc. as
                documented here:
                https://www.elastic.co/guide/en/elasticsearch/reference/master/sort-search-results.html

        Returns:
            A dataframe containing all matches as rows.
        """
        hits, next_cursor = self.get_n_samples_for_fields(search_term,
                                                          fields,
                                                          n,
                                                          cursor,
                                                          sort_by)
        search_df = json_normalize(hits)
        return search_df, next_cursor

    def get_all_samples_for_fields_as_df(self,
                                         search_term,
                                         fields=[],
                                         sort_by=Discover.DEFAULT_SORT_ORDER):
        """Obtain a dataframe of all sample matches against the `search_term`
        in any of the given `fields`.

        Args:
            search_term (str):
                Expression to be searched for. Must be a valid Elasticsearch
                "query" string.
            fields (list):
                Name of the fields whose values are to be matched against.
            sort_by (list):
                A list of dictionaries which will be used to order responses
                for queries. Defaults to sorting based on combination of score
                (descending) and ID (ascending).

                Each dictionary is a key-value pair, where the keys are fields
                to be sorted on and the values may contain order, mode, etc. as
                documented here:
                https://www.elastic.co/guide/en/elasticsearch/reference/master/sort-search-results.html

        Returns:
            A dataframe containing all matches as rows.
        """
        query_args = (search_term, fields)
        all_hits = self.all_matches_for_query(
            self.get_n_samples_for_fields, query_args, sort_by)
        search_df = json_normalize(all_hits)
        return search_df

    def get_n_samples_for_field_combinations(
            self,
            and_fields={},
            or_fields={},
            not_fields={},
            n=Discover.DEFAULT_PAGE_SIZE,
            cursor=None,
            sort_by=Discover.DEFAULT_SORT_ORDER):
        """Fetch sample entries that satisfy the given match combinations.

        Args:
            and_fields (dict):
                A dictionary of fields as keys mapping to their expected
                values. All fields within this dictionary must match.
            or_fields (dict):
                A dictionary of fields as keys mapping to their expected
                values. Any of the fields can be present in the match.
            not_fields (dict):
                A dictionary of fields as keys mapping to their expected
                values. None of the fields must match.
            n (int, optional):
                Number of matches to fetch. Defaults to
                `Discover.DEFAULT_PAGE_SIZE`.
            cursor (list, optional):
                A cursor from where to begin searching. Can be obtained from
                calls to this method. Defaults to None, in case search needs to
                start at the beginning.
            sort_by (list):
                A list of dictionaries which will be used to order responses
                for queries. Defaults to sorting based on combination of score
                (descending) and ID (ascending).

                Each dictionary is a key-value pair, where the keys are fields
                to be sorted on and the values may contain order, mode, etc. as
                documented here:
                https://www.elastic.co/guide/en/elasticsearch/reference/master/sort-search-results.html

        Returns:
            A list of hits and a cursor. The list contains the `n` hits as
            rows. The cursor can be used to fetch the next set of items from
            where this one ended.
        """
        and_match_list = [{"match": {k: v}} for k, v in and_fields.items()]
        or_match_list = [{"match": {k: v}} for k, v in or_fields.items()]
        not_match_list = [{"match": {k: v}} for k, v in not_fields.items()]

        payload = {
            "query": {
                "bool": {
                    "must": and_match_list,
                    "must_not": not_match_list,
                    "should": or_match_list
                }
            },
        }
        return self._execute_search_for_hits(payload, n, cursor, sort_by)

    def query_samples_by_field_combination(self,
                                           and_fields={},
                                           or_fields={},
                                           not_fields={},
                                           n=Discover.DEFAULT_PAGE_SIZE):
        print("Warning: this method has been deprecated. Please consider",
              "using 'get_n_samples_for_field_combinations_as_df' instead.")

        hits, _ = self.get_n_samples_for_field_combinations(and_fields,
                                                            or_fields,
                                                            not_fields,
                                                            n,
                                                            None)
        search_df = json_normalize(hits)
        return search_df

    def get_n_samples_for_field_combinations_as_df(
            self,
            and_fields={},
            or_fields={},
            not_fields={},
            n=Discover.DEFAULT_PAGE_SIZE,
            cursor=None,
            sort_by=Discover.DEFAULT_SORT_ORDER):
        """Obtain a dataframe of specified number of sample matches that
        satisfy the given match combinations.

        Args:
            and_fields (dict):
                A dictionary of fields as keys mapping to their expected
                values. All fields within this dictionary must match.
            or_fields (dict):
                A dictionary of fields as keys mapping to their expected
                values. Any of the fields can be present in the match.
            not_fields (dict):
                A dictionary of fields as keys mapping to their expected
                values. None of the fields must match.
            n (int, optional):
                Number of matches to fetch. Defaults to
                `Discover.DEFAULT_PAGE_SIZE`.
            cursor (list, optional):
                A cursor from where to begin searching. Can be obtained from
                calls to this method. Defaults to None, in case search needs to
                start at the beginning.
            sort_by (list):
                A list of dictionaries which will be used to order responses
                for queries. Defaults to sorting based on combination of score
                (descending) and ID (ascending).

                Each dictionary is a key-value pair, where the keys are fields
                to be sorted on and the values may contain order, mode, etc. as
                documented here:
                https://www.elastic.co/guide/en/elasticsearch/reference/master/sort-search-results.html

        Returns:
            A dataframe and a cursor. The dataframe contains the `n` hits as
            rows. The cursor can be used to fetch the next set of items from
            where this one ended.
        """
        hits, next_cursor = self.get_n_samples_for_field_combinations(
            and_fields, or_fields, not_fields, n, cursor, sort_by)
        search_df = json_normalize(hits)
        return search_df, next_cursor

    def get_all_samples_for_field_combinations_as_df(
            self,
            and_fields={},
            or_fields={},
            not_fields={},
            sort_by=Discover.DEFAULT_SORT_ORDER):
        """Obtain a dataframe of all samples that satisfy the given match
        combinations.

        Args:
            and_fields (dict):
                A dictionary of fields as keys mapping to their expected
                values. All fields within this dictionary must match.
            or_fields (dict):
                A dictionary of fields as keys mapping to their expected
                values. Any of the fields can be present in the match.
            not_fields (dict):
                A dictionary of fields as keys mapping to their expected
                values. None of the fields must match.
            sort_by (list):
                A list of dictionaries which will be used to order responses
                for queries. Defaults to sorting based on combination of score
                (descending) and ID (ascending).

                Each dictionary is a key-value pair, where the keys are fields
                to be sorted on and the values may contain order, mode, etc. as
                documented here:
                https://www.elastic.co/guide/en/elasticsearch/reference/master/sort-search-results.html

        Returns:
            A dataframe containing all matches as rows.
        """
        query_args = (and_fields, or_fields, not_fields)
        all_hits = self.all_matches_for_query(
            self.get_n_samples_for_field_combinations, query_args, sort_by)
        search_df = json_normalize(all_hits)
        return search_df
